import sys
import time
import amanobot
from amanobot.text import apply_entities_as_markdown, apply_entities_as_html

TOKEN = sys.argv[1]
USER_ID = long(sys.argv[2])

bot = amanobot.Bot(TOKEN)

user_link = 'tg://user?id=' + str(USER_ID)

markdowns = [
    '*abc*de',
    'ab_cd_e',
    '[abcde](http://www.yahoo.com/)',
    '[user]('+user_link+')',
    'a`bcd`e',
    '''Below is a function:
```text
def add(a, b):
    return a+b
```
Do you know what it does?''',
    'a_bc_de*f*g`hijk`lmno',
    'ab_cd_*efg*`h`[ijkl](http://www.yahoo.com/)',
    'a*bcdefg*h[user]('+user_link+')',
    '''Markdown examples:
*1.* \\*bold text\\*
_2._ \\_italic text\\_
3. \\[inline URL](http://www.example.com/)
`4.` \\`inline fixed-width code\\`''',
]

print 'Testing Markdown ...'
for s in markdowns:
    msg = bot.sendMessage(USER_ID, s, parse_mode='Markdown')

    u = apply_entities_as_markdown(msg['text'], msg['entities'])

    if s == u:
        print 'Identical'
    else:
        print 'Different:'
        print 'Original ->', s
        print 'Applied ->', u

    time.sleep(2)


htmls = [
    'a<b>bcd</b>e',
    '<i>ab</i>cde',
    'ab<a href="http://www.yahoo.com/">cde</a>',
    'ab<a href="'+user_link+'">user</a>',
    'a<code>bcd</code>e',
    '''Below is a function:
<pre>
def add(a, b):
    return a+b
</pre>
Do you know what it does?''',
    'a<i>bc</i>de<b>f</b>g<code>hijk</code>lmno',
    'ab<i>cd</i><b>efg</b><code>h</code><a href="http://www.yahoo.com/">ijkl</a>',
    'a<b>bcdefg</b>h<a href="'+user_link+'">user</a>',
    '''HTML examples:
<b>1.</b> &lt;b&gt;bold&lt;/b&gt;
<i>2.</i> &lt;i&gt;italic&lt;/i&gt;
3. &lt;a href="http://www.example.com/"&gt;inline URL&lt;/a&gt;
<code>4.</code> &lt;code&gt;inline fixed-width code&lt;/code&gt;''',
]

print 'Testing HTML ...'
for s in htmls:
    msg = bot.sendMessage(USER_ID, s, parse_mode='HTML')

    u = apply_entities_as_html(msg['text'], msg['entities'])

    if s == u:
        print 'Identical'
    else:
        print 'Different:'
        print 'Original ->', s
        print 'Applied ->', u

    time.sleep(2)



import traceback
from functools import wraps
from . import exception
from . import flavor, peel, is_event, chat_flavors, inline_flavors

def _wrap_none(fn):
    def w(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except (KeyError, exception.BadFlavor):
            return None
    return w

def per_chat_id(types='all'):
    """
    :param types:
        ``all`` or a list of chat types (``private``, ``group``, ``channel``)

    :return:
        a seeder function that returns the chat id only if the chat type is in ``types``.
    """
    return _wrap_none(lambda msg:
                          msg['chat']['id']
                          if types == 'all' or msg['chat']['type'] in types
                          else None)

def per_chat_id_in(s, types='all'):
    """
    :param s:
        a list or set of chat id

    :param types:
        ``all`` or a list of chat types (``private``, ``group``, ``channel``)

    :return:
        a seeder function that returns the chat id only if the chat id is in ``s``
        and chat type is in ``types``.
    """
    return _wrap_none(lambda msg:
                          msg['chat']['id']
                          if (types == 'all' or msg['chat']['type'] in types) and msg['chat']['id'] in s
                          else None)

def per_chat_id_except(s, types='all'):
    """
    :param s:
        a list or set of chat id

    :param types:
        ``all`` or a list of chat types (``private``, ``group``, ``channel``)

    :return:
        a seeder function that returns the chat id only if the chat id is *not* in ``s``
        and chat type is in ``types``.
    """
    return _wrap_none(lambda msg:
                          msg['chat']['id']
                          if (types == 'all' or msg['chat']['type'] in types) and msg['chat']['id'] not in s
                          else None)

def per_from_id(flavors=chat_flavors+inline_flavors):
    """
    :param flavors:
        ``all`` or a list of flavors

    :return:
        a seeder function that returns the from id only if the message flavor is
        in ``flavors``.
    """
    return _wrap_none(lambda msg:
                          msg['from']['id']
                          if flavors == 'all' or flavor(msg) in flavors
                          else None)

def per_from_id_in(s, flavors=chat_flavors+inline_flavors):
    """
    :param s:
        a list or set of from id

    :param flavors:
        ``all`` or a list of flavors

    :return:
        a seeder function that returns the from id only if the from id is in ``s``
        and message flavor is in ``flavors``.
    """
    return _wrap_none(lambda msg:
                          msg['from']['id']
                          if (flavors == 'all' or flavor(msg) in flavors) and msg['from']['id'] in s
                          else None)

def per_from_id_except(s, flavors=chat_flavors+inline_flavors):
    """
    :param s:
        a list or set of from id

    :param flavors:
        ``all`` or a list of flavors

    :return:
        a seeder function that returns the from id only if the from id is *not* in ``s``
        and message flavor is in ``flavors``.
    """
    return _wrap_none(lambda msg:
                          msg['from']['id']
                          if (flavors == 'all' or flavor(msg) in flavors) and msg['from']['id'] not in s
                          else None)

def per_inline_from_id():
    """
    :return:
        a seeder function that returns the from id only if the message flavor
        is ``inline_query`` or ``chosen_inline_result``
    """
    return per_from_id(flavors=inline_flavors)

def per_inline_from_id_in(s):
    """
    :param s: a list or set of from id
    :return:
        a seeder function that returns the from id only if the message flavor
        is ``inline_query`` or ``chosen_inline_result`` and the from id is in ``s``.
    """
    return per_from_id_in(s, flavors=inline_flavors)

def per_inline_from_id_except(s):
    """
    :param s: a list or set of from id
    :return:
        a seeder function that returns the from id only if the message flavor
        is ``inline_query`` or ``chosen_inline_result`` and the from id is *not* in ``s``.
    """
    return per_from_id_except(s, flavors=inline_flavors)

def per_application():
    """
    :return:
        a seeder function that always returns 1, ensuring at most one delegate is ever spawned
        for the entire application.
    """
    return lambda msg: 1

def per_message(flavors='all'):
    """
    :param flavors: ``all`` or a list of flavors
    :return:
        a seeder function that returns a non-hashable only if the message flavor
        is in ``flavors``.
    """
    return _wrap_none(lambda msg: [] if flavors == 'all' or flavor(msg) in flavors else None)

def per_event_source_id(event_space):
    """
    :return:
        a seeder function that returns an event's source id only if that event's
        source space equals to ``event_space``.
    """
    def f(event):
        if is_event(event):
            v = peel(event)
            if v['source']['space'] == event_space:
                return v['source']['id']
            else:
                return None
        else:
            return None
    return _wrap_none(f)

def per_callback_query_chat_id(types='all'):
    """
    :param types:
        ``all`` or a list of chat types (``private``, ``group``, ``channel``)

    :return:
        a seeder function that returns a callback query's originating chat id
        if the chat type is in ``types``.
    """
    def f(msg):
        if (flavor(msg) == 'callback_query' and 'message' in msg
            and (types == 'all' or msg['message']['chat']['type'] in types)):
            return msg['message']['chat']['id']
        else:
            return None
    return f

def per_callback_query_origin(origins='all'):
    """
    :param origins:
        ``all`` or a list of origin types (``chat``, ``inline``)

    :return:
        a seeder function that returns a callback query's origin identifier if
        that origin type is in ``origins``. The origin identifier is guaranteed
        to be a tuple.
    """
    def f(msg):
        def origin_type_ok():
            return (origins == 'all'
                or ('chat' in origins and 'message' in msg)
                or ('inline' in origins and 'inline_message_id' in msg))

        if flavor(msg) == 'callback_query' and origin_type_ok():
            if 'inline_message_id' in msg:
                return msg['inline_message_id'],
            else:
                return msg['message']['chat']['id'], msg['message']['message_id']
        else:
            return None
    return f

def per_invoice_payload():
    """
    :return:
        a seeder function that returns the invoice payload.
    """
    def f(msg):
        if 'successful_payment' in msg:
            return msg['successful_payment']['invoice_payload']
        else:
            return msg['invoice_payload']

    return _wrap_none(f)

def call(func, *args, **kwargs):
    """
    :return:
        a delegator function that returns a tuple (``func``, (seed tuple,)+ ``args``, ``kwargs``).
        That is, seed tuple is inserted before supplied positional arguments.
        By default, a thread wrapping ``func`` and all those arguments is spawned.
    """
    def f(seed_tuple):
        return func, (seed_tuple,)+args, kwargs
    return f

def create_run(cls, *args, **kwargs):
    """
    :return:
        a delegator function that calls the ``cls`` constructor whose arguments being
        a seed tuple followed by supplied ``*args`` and ``**kwargs``, then returns
        the object's ``run`` method. By default, a thread wrapping that ``run`` method
        is spawned.
    """
    def f(seed_tuple):
        j = cls(seed_tuple, *args, **kwargs)
        return j.run
    return f

def create_open(cls, *args, **kwargs):
    """
    :return:
        a delegator function that calls the ``cls`` constructor whose arguments being
        a seed tuple followed by supplied ``*args`` and ``**kwargs``, then returns
        a looping function that uses the object's ``listener`` to wait for messages
        and invokes instance method ``open``, ``on_message``, and ``on_close`` accordingly.
        By default, a thread wrapping that looping function is spawned.
    """
    def f(seed_tuple):
        j = cls(seed_tuple, *args, **kwargs)

        def wait_loop():
            bot, msg, seed = seed_tuple
            try:
                handled = j.open(msg, seed)
                if not handled:
                    j.on_message(msg)

                while 1:
                    msg = j.listener.wait()
                    j.on_message(msg)

            # These exceptions are "normal" exits.
            except (exception.IdleTerminate, exception.StopListening) as e:
                j.on_close(e)

            # Any other exceptions are accidents. **Print it out.**
            # This is to prevent swallowing exceptions in the case that on_close()
            # gets overridden but fails to account for unexpected exceptions.
            except Exception as e:
                traceback.print_exc()
                j.on_close(e)

        return wait_loop
    return f

def until(condition, fns):
    """
    Try a list of seeder functions until a condition is met.

    :param condition:
        a function that takes one argument - a seed - and returns ``True``
        or ``False``

    :param fns:
        a list of seeder functions

    :return:
        a "composite" seeder function that calls each supplied function in turn,
        and returns the first seed where the condition is met. If the condition
        is never met, it returns ``None``.
    """
    def f(msg):
        for fn in fns:
            seed = fn(msg)
            if condition(seed):
                return seed
        return None
    return f

def chain(*fns):
    """
    :return:
        a "composite" seeder function that calls each supplied function in turn,
        and returns the first seed that is not ``None``.
    """
    return until(lambda seed: seed is not None, fns)

def _ensure_seeders_list(fn):
    @wraps(fn)
    def e(seeders, *aa, **kw):
        return fn(seeders if isinstance(seeders, list) else [seeders], *aa, **kw)
    return e

@_ensure_seeders_list
def pair(seeders, delegator_factory, *args, **kwargs):
    """
    The basic pair producer.

    :return:
        a (seeder, delegator_factory(\*args, \*\*kwargs)) tuple.

    :param seeders:
        If it is a seeder function or a list of one seeder function, it is returned
        as the final seeder. If it is a list of more than one seeder function, they
        are chained together before returned as the final seeder.
    """
    return (chain(*seeders) if len(seeders) > 1 else seeders[0],
            delegator_factory(*args, **kwargs))

def _natural_numbers():
    x = 0
    while 1:
        x += 1
        yield x

_event_space = _natural_numbers()

def pave_event_space(fn=pair):
    """
    :return:
        a pair producer that ensures the seeder and delegator share the same event space.
    """
    global _event_space
    event_space = next(_event_space)

    @_ensure_seeders_list
    def p(seeders, delegator_factory, *args, **kwargs):
        return fn(seeders + [per_event_source_id(event_space)],
                  delegator_factory, *args, event_space=event_space, **kwargs)
    return p

def include_callback_query_chat_id(fn=pair, types='all'):
    """
    :return:
        a pair producer that enables static callback query capturing
        across seeder and delegator.

    :param types:
        ``all`` or a list of chat types (``private``, ``group``, ``channel``)
    """
    @_ensure_seeders_list
    def p(seeders, delegator_factory, *args, **kwargs):
        return fn(seeders + [per_callback_query_chat_id(types=types)],
                  delegator_factory, *args, include_callback_query=True, **kwargs)
    return p

from . import helper

def intercept_callback_query_origin(fn=pair, origins='all'):
    """
    :return:
        a pair producer that enables dynamic callback query origin mapping
        across seeder and delegator.

    :param origins:
        ``all`` or a list of origin types (``chat``, ``inline``).
        Origin mapping is only enabled for specified origin types.
    """
    origin_map = helper.SafeDict()

    # For key functions that returns a tuple as key (e.g. per_callback_query_origin()),
    # wrap the key in another tuple to prevent router from mistaking it as
    # a key followed by some arguments.
    def tuplize(fn):
        def tp(msg):
            return (fn(msg),)
        return tp

    router = helper.Router(tuplize(per_callback_query_origin(origins=origins)),
                           origin_map)

    def modify_origin_map(origin, dest, set):
        if set:
            origin_map[origin] = dest
        else:
            try:
                del origin_map[origin]
            except KeyError:
                pass

    if origins == 'all':
        intercept = modify_origin_map
    else:
        intercept = (modify_origin_map if 'chat' in origins else False,
                     modify_origin_map if 'inline' in origins else False)

    @_ensure_seeders_list
    def p(seeders, delegator_factory, *args, **kwargs):
        return fn(seeders + [_wrap_none(router.map)],
                  delegator_factory, *args, intercept_callback_query=intercept, **kwargs)
    return p



import time
import amanobot
from amanobot.loop import OrderedWebhook

def u(update_id):
    return { 'update_id': update_id, 'message': update_id }

sequence = [
    u(1),  # initialize
    u(2),  # no buffering

    u(4),  # 1-gap
    u(3),  # clear 2

    u(7),  # 2-gap
    u(5),  # return, leave 1-gap
    u(6),  # clear 2

    u(10),  # 2-gap
    u(9),   # 1-gap
    u(8),   # clear 3

    u(15),
    u(12),
    u(13),
    u(11),
    u(14),

    u(17),
    u(18),
    u(21),
    u(20),
    u(19),
    u(16),

    u(22),  # no buffering

    u(24),
    9,  # skip id=23

    u(23),  # discard

    u(26),
    u(27),
    9,  # skip id=25

    u(25),  # discard

    u(30),
    u(29),
    5,
    u(32),
    u(33),
    2,  # clear 29,30, skip 28
    u(31),  # clear 31,32,33

    u(39),
    u(36),
    2,
    u(37),
    7,  # clear 36,37,39

    u(28),  # discard
    u(38),  # discard

    u(40),  # return
]

def handle(msg):
    print msg

bot = amanobot.Bot('abc')
webhook = OrderedWebhook(bot, handle)

webhook.run_as_thread(maxhold=8)

for update in sequence:
    if type(update) is dict:
        webhook.feed(update)
        time.sleep(1)
    else:
        time.sleep(update)



# coding=utf8

import time
import threading
import pprint
import sys
import traceback
import random
import amanobot
from amanobot.namedtuple import (
    InlineQuery, ChosenInlineResult, InputTextMessageContent,
    InlineQueryResultArticle, InlineQueryResultPhoto, InlineQueryResultGame)

def equivalent(data, nt):
    if type(data) is dict:
        keys = data.keys()

        # number of dictionary keys == number of non-None values in namedtuple?
        if len(keys) != len([f for f in nt._fields if getattr(nt, f) is not None]):
            return False

        # map `from` to `from_`
        fields = list(map(lambda k: k+'_' if k in ['from'] else k, keys))

        return all(map(equivalent, [data[k] for k in keys], [getattr(nt, f) for f in fields]))
    elif type(data) is list:
        return all(map(equivalent, data, nt))
    else:
        return data==nt

def examine(result, type):
    try:
        print 'Examining %s ......' % type

        nt = type(**result)
        assert equivalent(result, nt), 'Not equivalent:::::::::::::::\n%s\n::::::::::::::::\n%s' % (result, nt)

        pprint.pprint(result)
        pprint.pprint(nt)
        print
    except AssertionError:
        traceback.print_exc()
        answer = raw_input('Do you want to continue? [y] ')
        if answer != 'y':
            exit(1)


def on_inline_query(msg):
    def compute():
        articles = [InlineQueryResultArticle(
                       id='abc', title='HK', input_message_content=InputTextMessageContent(message_text='Hong Kong'), url='https://www.google.com', hide_url=True),
                   {'type': 'article',
                       'id': 'def', 'title': 'SZ', 'input_message_content': {'message_text': 'Shenzhen'}, 'url': 'https://www.yahoo.com'}]

        photos = [InlineQueryResultPhoto(
                      id='123', photo_url='https://core.telegram.org/file/811140934/1/tbDSLHSaijc/fdcc7b6d5fb3354adf', thumb_url='https://core.telegram.org/file/811140934/1/tbDSLHSaijc/fdcc7b6d5fb3354adf'),
                  {'type': 'photo',
                      'id': '345', 'photo_url': 'https://core.telegram.org/file/811140184/1/5YJxx-rostA/ad3f74094485fb97bd', 'thumb_url': 'https://core.telegram.org/file/811140184/1/5YJxx-rostA/ad3f74094485fb97bd', 'caption': 'Caption', 'title': 'Title', 'input_message_content': {'message_text': 'Shenzhen'}}]

        games = [InlineQueryResultGame(
                    id='abc', game_short_name='sunchaser')]

        results = random.choice([articles, photos, games])
        return results

    query_id, from_id, query = amanobot.glance(msg, flavor='inline_query')

    if from_id != USER_ID:
        print 'Unauthorized user:', from_id
        return

    examine(msg, InlineQuery)
    answerer.answer(msg, compute)


def on_chosen_inline_result(msg):
    result_id, from_id, query = amanobot.glance(msg, flavor='chosen_inline_result')

    if from_id != USER_ID:
        print 'Unauthorized user:', from_id
        return

    examine(msg, ChosenInlineResult)

    print 'Chosen inline query:'
    pprint.pprint(msg)


TOKEN = sys.argv[1]
USER_ID = long(sys.argv[2])

bot = amanobot.Bot(TOKEN)
answerer = amanobot.helper.Answerer(bot)

bot.sendMessage(USER_ID, 'Please give me an inline query.')

bot.message_loop({'inline_query': on_inline_query,
                  'chosen_inline_result': on_chosen_inline_result}, run_forever=True)



import collections
import warnings
import sys


class _Field(object):
    def __init__(self, name, constructor=None, default=None):
        self.name = name
        self.constructor = constructor
        self.default = default


# Function to produce namedtuple classes.
def _create_class(typename, fields):
    # extract field names
    field_names = [e.name if type(e) is _Field else e for e in fields]

    # Some dictionary keys are Python keywords and cannot be used as field names, e.g. `from`.
    # Get around by appending a '_', e.g. dict['from'] => namedtuple.from_
    keymap = [(k.rstrip('_'), k) for k in filter(lambda e: e in ['from_'], field_names)]

    # extract (non-simple) fields that need conversions
    conversions = [(e.name, e.constructor) for e in fields if type(e) is _Field and e.constructor is not None]

    # extract default values
    defaults = [e.default if type(e) is _Field else None for e in fields]

    # Create the base tuple class, with defaults.
    base = collections.namedtuple(typename, field_names)
    base.__new__.__defaults__ = tuple(defaults)

    class sub(base):
        def __new__(cls, **kwargs):
            # Map keys.
            for oldkey, newkey in keymap:
                if oldkey in kwargs:
                    kwargs[newkey] = kwargs[oldkey]
                    del kwargs[oldkey]

            # Any unexpected arguments?
            unexpected = set(kwargs.keys()) - set(super(sub, cls)._fields)

            # Remove unexpected arguments and issue warning.
            if unexpected:
                for k in unexpected:
                    del kwargs[k]

                s = ('Unexpected fields: ' + ', '.join(unexpected) + ''
                     '\nBot API seems to have added new fields to the returned data.'
                     ' This version of namedtuple is not able to capture them.'
                     '\n\nPlease upgrade amanobot by:'
                     '\n  sudo pip install amanobot --upgrade'
                     '\n\nIf you still see this message after upgrade, that means I am still working to bring the code'
                     ' up-to-date. Please try upgrade again a few days later. In the meantime,'
                     ' you can access the new fields the old-fashioned way, through the raw dictionary.')

                warnings.warn(s, UserWarning)

            # Convert non-simple values to namedtuples.
            for key, func in conversions:
                if key in kwargs:
                    if type(kwargs[key]) is dict:
                        kwargs[key] = func(**kwargs[key])
                    elif type(kwargs[key]) is list:
                        kwargs[key] = func(kwargs[key])
                    else:
                        raise RuntimeError('Can only convert dict or list')

            return super(sub, cls).__new__(cls, **kwargs)

    # https://bugs.python.org/issue24931
    # Python 3.4 bug: namedtuple subclass does not inherit __dict__ properly.
    # Fix it manually.
    if sys.version_info >= (3, 4):
        def _asdict(self):
            return collections.OrderedDict(zip(self._fields, self))
        sub._asdict = _asdict

    sub.__name__ = typename

    return sub

"""
Different treatments for incoming and outgoing namedtuples:

- Incoming ones require type declarations for certain fields for deeper parsing.
- Outgoing ones need no such declarations because users are expected to put the correct object in place.
"""


# Namedtuple class will reference other namedtuple classes. Due to circular
# dependencies, it is impossible to have all class definitions ready at
# compile time. We have to dynamically obtain class reference at runtime.
# For example, the following function acts like a constructor for `Message`
# so any class can reference the Message namedtuple even before the Message
# namedtuple is defined.
def _Message(**kwargs):
    return getattr(sys.modules[__name__], 'Message')(**kwargs)


# incoming
User = _create_class('User', [
           'id',
           'is_bot',
           'first_name',
           'last_name',
           'username',
           'language_code'
       ])


def UserArray(data):
    return [User(**p) for p in data]


# incoming
ChatPhoto = _create_class('ChatPhoto', [
                'small_file_id',
                'big_file_id',
            ])

# incoming
ChatPermissions = _create_class('ChatPermissions', [
                      'can_send_messages',
                      'can_send_media_messages',
                      'can_send_polls',
                      'can_send_other_messages',
                      'can_add_web_page_previews',
                      'can_change_info',
                      'can_invite_users',
                      'can_pin_messages'
])

# incoming
Chat = _create_class('Chat', [
           'id',
           'type',
           'title',
           'username',
           'first_name',
           'last_name',
           _Field('photo', constructor=ChatPhoto),
           'description',
           'invite_link',
           _Field('pinned_message', constructor=_Message),
           _Field('permissions', constructor=ChatPermissions),
           'sticker_set_name',
           'can_set_sticker_set',
       ])

# incoming
PhotoSize = _create_class('PhotoSize', [
                'file_id',
                'width',
                'height',
                'file_size',
                'file_path',  # undocumented
            ])

# incoming
Audio = _create_class('Audio', [
            'file_id',
            'duration',
            'performer',
            'title',
            'mime_type',
            'file_size',
            _Field('thumb', constructor=PhotoSize)
        ])

# incoming
Document = _create_class('Document', [
               'file_id',
               _Field('thumb', constructor=PhotoSize),
               'file_name',
               'mime_type',
               'file_size',
               'file_path',  # undocumented
           ])

# incoming and outgoing
MaskPosition = _create_class('MaskPosition', [
                   'point',
                   'x_shift',
                   'y_shift',
                   'scale',
               ])

# incoming
Sticker = _create_class('Sticker', [
              'file_id',
              'width',
              'height',
              'is_animated',
              _Field('thumb', constructor=PhotoSize),
              'emoji',
              'set_name',
              _Field('mask_position', constructor=MaskPosition),
              'file_size',
          ])


def StickerArray(data):
    return [Sticker(**p) for p in data]


# incoming
StickerSet = _create_class('StickerSet', [
                 'name',
                 'title',
                 'is_animated',
                 'contains_masks',
                 _Field('stickers', constructor=StickerArray),
             ])

# incoming
Video = _create_class('Video', [
            'file_id',
            'width',
            'height',
            'duration',
            _Field('thumb', constructor=PhotoSize),
            'mime_type',
            'file_size',
            'file_path',  # undocumented
        ])

# incoming
Voice = _create_class('Voice', [
            'file_id',
            'duration',
            'mime_type',
            'file_size'
        ])

# incoming
VideoNote = _create_class('VideoNote', [
                'file_id',
                'length',
                'duration',
                _Field('thumb', constructor=PhotoSize),
                'file_size'
            ])

# incoming
Contact = _create_class('Contact', [
              'phone_number',
              'first_name',
              'last_name',
              'user_id',
              'vcard'
          ])

# incoming
Location = _create_class('Location', [
               'longitude',
               'latitude'
           ])

# incoming
Venue = _create_class('Venue', [
            _Field('location', constructor=Location),
            'title',
            'address',
            'foursquare_id',
        ])

# incoming
PollOption = _create_class('PollOption', [
                 'text',
                 'voter_count'
             ])

# incoming
Poll = _create_class('Poll', [
           'id',
           'question',
           'options',
           'is_closed'
       ])

# incoming
File = _create_class('File', [
           'file_id',
           'file_size',
           'file_path'
       ])


def PhotoSizeArray(data):
    return [PhotoSize(**p) for p in data]


def PhotoSizeArrayArray(data):
    return [[PhotoSize(**p) for p in array] for array in data]


# incoming
UserProfilePhotos = _create_class('UserProfilePhotos', [
                        'total_count',
                        _Field('photos', constructor=PhotoSizeArrayArray)
                    ])

# incoming
ChatMember = _create_class('ChatMember', [
                 _Field('user', constructor=User),
                 'status',
                 'until_date',
                 'can_be_edited',
                 'can_post_messages',
                 'can_edit_messages',
                 'can_delete_messages',
                 'can_restrict_members',
                 'can_promote_members',
                 'can_change_info',
                 'can_invite_users',
                 'can_pin_messages',
                 'is_member',
                 'can_send_messages',
                 'can_send_media_messages',
                 'can_send_polls',
                 'can_send_other_messages',
                 'can_add_web_page_previews',
             ])


def ChatMemberArray(data):
    return [ChatMember(**p) for p in data]


# outgoing
ReplyKeyboardMarkup = _create_class('ReplyKeyboardMarkup', [
                          'keyboard',
                          'resize_keyboard',
                          'one_time_keyboard',
                          'selective',
                      ])

# outgoing
KeyboardButton = _create_class('KeyboardButton', [
                     'text',
                     'request_contact',
                     'request_location',
                 ])

# outgoing
LoginUrl = _create_class('LoginUrl', [
               'url',
               'forward_text',
               'bot_username',
               'request_write_access'
           ])

# outgoing
ReplyKeyboardRemove = _create_class('ReplyKeyboardRemove', [
                          _Field('remove_keyboard', default=True),
                          'selective',
                      ])

# outgoing
ForceReply = _create_class('ForceReply', [
                 _Field('force_reply', default=True),
                 'selective',
             ])

# outgoing
InlineKeyboardButton = _create_class('InlineKeyboardButton', [
                           'text',
                           'url',
                           _Field('login_url', constructor=LoginUrl),
                           'callback_data',
                           'switch_inline_query',
                           'switch_inline_query_current_chat',
                           'callback_game',
                           'pay',
                       ])

# outgoing
InlineKeyboardMarkup = _create_class('InlineKeyboardMarkup', [
                           'inline_keyboard',
                       ])

# incoming
MessageEntity = _create_class('MessageEntity', [
                    'type',
                    'offset',
                    'length',
                    'url',
                    _Field('user', constructor=User),
                ])


# incoming
def MessageEntityArray(data):
    return [MessageEntity(**p) for p in data]


# incoming
GameHighScore = _create_class('GameHighScore', [
                    'position',
                    _Field('user', constructor=User),
                    'score',
                ])

# incoming
Animation = _create_class('Animation', [
                'file_id',
                _Field('thumb', constructor=PhotoSize),
                'file_name',
                'mime_type',
                'file_size',
            ])

# incoming
Game = _create_class('Game', [
           'title',
           'description',
           _Field('photo', constructor=PhotoSizeArray),
           'text',
           _Field('text_entities', constructor=MessageEntityArray),
           _Field('animation', constructor=Animation),
       ])

# incoming
Invoice = _create_class('Invoice', [
              'title',
              'description',
              'start_parameter',
              'currency',
              'total_amount',
          ])

# outgoing
LabeledPrice = _create_class('LabeledPrice', [
                   'label',
                   'amount',
               ])

# outgoing
ShippingOption = _create_class('ShippingOption', [
                     'id',
                     'title',
                     'prices',
                 ])

# incoming
ShippingAddress = _create_class('ShippingAddress', [
                      'country_code',
                      'state',
                      'city',
                      'street_line1',
                      'street_line2',
                      'post_code',
                  ])

# incoming
OrderInfo = _create_class('OrderInfo', [
                'name',
                'phone_number',
                'email',
                _Field('shipping_address', constructor=ShippingAddress),
            ])

# incoming
ShippingQuery = _create_class('ShippingQuery', [
                    'id',
                    _Field('from_', constructor=User),
                    'invoice_payload',
                    _Field('shipping_address', constructor=ShippingAddress),
                ])

# incoming
PreCheckoutQuery = _create_class('PreCheckoutQuery', [
                       'id',
                       _Field('from_', constructor=User),
                       'currency',
                       'total_amount',
                       'invoice_payload',
                       'shipping_option_id',
                       _Field('order_info', constructor=OrderInfo),
                   ])

# incoming
SuccessfulPayment = _create_class('SuccessfulPayment', [
                        'currency',
                        'total_amount',
                        'invoice_payload',
                        'shipping_option_id',
                        _Field('order_info', constructor=OrderInfo),
                        'telegram_payment_charge_id',
                        'provider_payment_charge_id',
                    ])

# outgoing
PassportFile = _create_class('PassportFile', [
                   'file_id',
                   'file_size',
                   'file_date'
               ])

# outgoing
EncryptedPassportElement = _create_class('EncryptedPassportElement', [
                               'type',
                               'data',
                               'phone_number',
                               'email',
                               'files',
                               _Field('front_side', constructor=PassportFile),
                               _Field('reverse_side', constructor=PassportFile),
                               _Field('selfie', constructor=PassportFile),
                               'translation',
                               'hash'
                           ])

# outgoing
EncryptedCredentials = _create_class('EncryptedCredentials', [
                           'data',
                           'hash',
                           'secret'
                       ])

# outgoing
PassportData = _create_class('PassportData', [
                   'data',
                   _Field('credentials', constructor=EncryptedCredentials)
               ])

# incoming
Message = _create_class('Message', [
              'message_id',
              _Field('from_', constructor=User),
              'date',
              _Field('chat', constructor=Chat),
              _Field('forward_from', constructor=User),
              _Field('forward_from_chat', constructor=Chat),
              'forward_from_message_id',
              'forward_signature',
              'forward_sender_name',
              'forward_date',
              _Field('reply_to_message', constructor=_Message),
              'edit_date',
              'author_signature',
              'text',
              _Field('entities', constructor=MessageEntityArray),
              _Field('caption_entities', constructor=MessageEntityArray),
              _Field('audio', constructor=Audio),
              _Field('document', constructor=Document),
              _Field('game', constructor=Game),
              _Field('photo', constructor=PhotoSizeArray),
              _Field('sticker', constructor=Sticker),
              _Field('video', constructor=Video),
              _Field('voice', constructor=Voice),
              _Field('video_note', constructor=VideoNote),
              _Field('new_chat_members', constructor=UserArray),
              'caption',
              _Field('contact', constructor=Contact),
              _Field('location', constructor=Location),
              _Field('venue', constructor=Venue),
              _Field('poll', constructor=Poll),
              _Field('new_chat_member', constructor=User),
              _Field('left_chat_member', constructor=User),
              'new_chat_title',
              _Field('new_chat_photo', constructor=PhotoSizeArray),
              'delete_chat_photo',
              'group_chat_created',
              'supergroup_chat_created',
              'channel_chat_created',
              'migrate_to_chat_id',
              'migrate_from_chat_id',
              _Field('pinned_message', constructor=_Message),
              _Field('invoice', constructor=Invoice),
              _Field('successful_payment', constructor=SuccessfulPayment),
              'connected_website',
              _Field('passport_data', constructor=PassportData),
              _Field('reply_markup', constructor=InlineKeyboardMarkup)
          ])

# incoming
InlineQuery = _create_class('InlineQuery', [
                  'id',
                  _Field('from_', constructor=User),
                  _Field('location', constructor=Location),
                  'query',
                  'offset',
              ])

# incoming
ChosenInlineResult = _create_class('ChosenInlineResult', [
                         'result_id',
                         _Field('from_', constructor=User),
                         _Field('location', constructor=Location),
                         'inline_message_id',
                         'query',
                     ])

# incoming
CallbackQuery = _create_class('CallbackQuery', [
                    'id',
                    _Field('from_', constructor=User),
                    _Field('message', constructor=Message),
                    'inline_message_id',
                    'chat_instance',
                    'data',
                    'game_short_name',
                ])

# incoming
Update = _create_class('Update', [
             'update_id',
             _Field('message', constructor=Message),
             _Field('edited_message', constructor=Message),
             _Field('channel_post', constructor=Message),
             _Field('edited_channel_post', constructor=Message),
             _Field('inline_query', constructor=InlineQuery),
             _Field('chosen_inline_result', constructor=ChosenInlineResult),
             _Field('callback_query', constructor=CallbackQuery),
             _Field('poll', constructor=Poll)
         ])


# incoming
def UpdateArray(data):
    return [Update(**u) for u in data]


# incoming
WebhookInfo = _create_class('WebhookInfo', [
                  'url',
                  'has_custom_certificate',
                  'pending_update_count',
                  'last_error_date',
                  'last_error_message',
              ])

# outgoing
InputTextMessageContent = _create_class('InputTextMessageContent', [
                              'message_text',
                              'parse_mode',
                              'disable_web_page_preview',
                          ])

# outgoing
InputLocationMessageContent = _create_class('InputLocationMessageContent', [
                                  'latitude',
                                  'longitude',
                                  'live_period',
                              ])

# outgoing
InputVenueMessageContent = _create_class('InputVenueMessageContent', [
                               'latitude',
                               'longitude',
                               'title',
                               'address',
                               'foursquare_id',
                               'foursquare_type'
                           ])

# outgoing
InputContactMessageContent = _create_class('InputContactMessageContent', [
                                 'phone_number',
                                 'first_name',
                                 'last_name',
                                 'vcard'
                             ])

# outgoing
InlineQueryResultArticle = _create_class('InlineQueryResultArticle', [
                               _Field('type', default='article'),
                               'id',
                               'title',
                               'input_message_content',
                               'reply_markup',
                               'url',
                               'hide_url',
                               'description',
                               'thumb_url',
                               'thumb_width',
                               'thumb_height',
                           ])

# outgoing
InlineQueryResultPhoto = _create_class('InlineQueryResultPhoto', [
                             _Field('type', default='photo'),
                             'id',
                             'photo_url',
                             'thumb_url',
                             'photo_width',
                             'photo_height',
                             'title',
                             'description',
                             'caption',
                             'parse_mode',
                             'reply_markup',
                             'input_message_content',
                         ])

# outgoing
InlineQueryResultGif = _create_class('InlineQueryResultGif', [
                           _Field('type', default='gif'),
                           'id',
                           'gif_url',
                           'gif_width',
                           'gif_height',
                           'gif_duration',
                           'thumb_url',
                           'title',
                           'caption',
                           'parse_mode',
                           'reply_markup',
                           'input_message_content',
                       ])

# outgoing
InlineQueryResultMpeg4Gif = _create_class('InlineQueryResultMpeg4Gif', [
                                _Field('type', default='mpeg4_gif'),
                                'id',
                                'mpeg4_url',
                                'mpeg4_width',
                                'mpeg4_height',
                                'mpeg4_duration',
                                'thumb_url',
                                'title',
                                'caption',
                                'parse_mode',
                                'reply_markup',
                                'input_message_content',
                            ])

# outgoing
InlineQueryResultVideo = _create_class('InlineQueryResultVideo', [
                             _Field('type', default='video'),
                             'id',
                             'video_url',
                             'mime_type',
                             'thumb_url',
                             'title',
                             'caption',
                             'parse_mode',
                             'video_width',
                             'video_height',
                             'video_duration',
                             'description',
                             'reply_markup',
                             'input_message_content',
                         ])

# outgoing
InlineQueryResultAudio = _create_class('InlineQueryResultAudio', [
                             _Field('type', default='audio'),
                             'id',
                             'audio_url',
                             'title',
                             'caption',
                             'parse_mode',
                             'performer',
                             'audio_duration',
                             'reply_markup',
                             'input_message_content',
                         ])

# outgoing
InlineQueryResultVoice = _create_class('InlineQueryResultVoice', [
                             _Field('type', default='voice'),
                             'id',
                             'voice_url',
                             'title',
                             'caption',
                             'parse_mode',
                             'voice_duration',
                             'reply_markup',
                             'input_message_content',
                         ])

# outgoing
InlineQueryResultDocument = _create_class('InlineQueryResultDocument', [
                                _Field('type', default='document'),
                                'id',
                                'title',
                                'caption',
                                'parse_mode',
                                'document_url',
                                'mime_type',
                                'description',
                                'reply_markup',
                                'input_message_content',
                                'thumb_url',
                                'thumb_width',
                                'thumb_height',
                            ])

# outgoing
InlineQueryResultLocation = _create_class('InlineQueryResultLocation', [
                                _Field('type', default='location'),
                                'id',
                                'latitude',
                                'longitude',
                                'title',
                                'live_period',
                                'reply_markup',
                                'input_message_content',
                                'thumb_url',
                                'thumb_width',
                                'thumb_height',
                            ])

# outgoing
InlineQueryResultVenue = _create_class('InlineQueryResultVenue', [
                                _Field('type', default='venue'),
                                'id',
                                'latitude',
                                'longitude',
                                'title',
                                'address',
                                'foursquare_id',
                                'foursquare_type',
                                'reply_markup',
                                'input_message_content',
                                'thumb_url',
                                'thumb_width',
                                'thumb_height',
                         ])

# outgoing
InlineQueryResultContact = _create_class('InlineQueryResultContact', [
                                _Field('type', default='contact'),
                                'id',
                                'phone_number',
                                'first_name',
                                'last_name',
                                'vcard',
                                'reply_markup',
                                'input_message_content',
                                'thumb_url',
                                'thumb_width',
                                'thumb_height',
                           ])

# outgoing
InlineQueryResultGame = _create_class('InlineQueryResultGame', [
                            _Field('type', default='game'),
                            'id',
                            'game_short_name',
                            'reply_markup',
                        ])

# outgoing
InlineQueryResultCachedPhoto = _create_class('InlineQueryResultCachedPhoto', [
                                   _Field('type', default='photo'),
                                   'id',
                                   'photo_file_id',
                                   'title',
                                   'description',
                                   'caption',
                                   'parse_mode',
                                   'reply_markup',
                                   'input_message_content',
                               ])

# outgoing
InlineQueryResultCachedGif = _create_class('InlineQueryResultCachedGif', [
                                 _Field('type', default='gif'),
                                 'id',
                                 'gif_file_id',
                                 'title',
                                 'caption',
                                 'parse_mode',
                                 'reply_markup',
                                 'input_message_content',
                             ])

# outgoing
InlineQueryResultCachedMpeg4Gif = _create_class('InlineQueryResultCachedMpeg4Gif', [
                                      _Field('type', default='mpeg4_gif'),
                                      'id',
                                      'mpeg4_file_id',
                                      'title',
                                      'caption',
                                      'parse_mode',
                                      'reply_markup',
                                      'input_message_content',
                                  ])

# outgoing
InlineQueryResultCachedSticker = _create_class('InlineQueryResultCachedSticker', [
                                     _Field('type', default='sticker'),
                                     'id',
                                     'sticker_file_id',
                                     'reply_markup',
                                     'input_message_content',
                                 ])

# outgoing
InlineQueryResultCachedDocument = _create_class('InlineQueryResultCachedDocument', [
                                      _Field('type', default='document'),
                                      'id',
                                      'title',
                                      'document_file_id',
                                      'description',
                                      'caption',
                                      'parse_mode',
                                      'reply_markup',
                                      'input_message_content',
                                  ])

# outgoing
InlineQueryResultCachedVideo = _create_class('InlineQueryResultCachedVideo', [
                                   _Field('type', default='video'),
                                   'id',
                                   'video_file_id',
                                   'title',
                                   'description',
                                   'caption',
                                   'parse_mode',
                                   'reply_markup',
                                   'input_message_content',
                               ])

# outgoing
InlineQueryResultCachedVoice = _create_class('InlineQueryResultCachedVoice', [
                                   _Field('type', default='voice'),
                                   'id',
                                   'voice_file_id',
                                   'title',
                                   'caption',
                                   'parse_mode',
                                   'reply_markup',
                                   'input_message_content',
                               ])

# outgoing
InlineQueryResultCachedAudio = _create_class('InlineQueryResultCachedAudio', [
                                   _Field('type', default='audio'),
                                   'id',
                                   'audio_file_id',
                                   'caption',
                                   'parse_mode',
                                   'reply_markup',
                                   'input_message_content',
                               ])

# outgoing
InputMediaPhoto = _create_class('InputMediaPhoto', [
                      _Field('type', default='photo'),
                      'media',
                      'thumb',
                      'caption',
                      'parse_mode',
                  ])

# outgoing
InputMediaVideo = _create_class('InputMediaVideo', [
                      _Field('type', default='video'),
                      'media',
                      'thumb',
                      'caption',
                      'parse_mode',
                      'width',
                      'height',
                      'duration',
                      'supports_streaming',
                  ])

# outgoing
InputMediaAnimation = _create_class('InputMediaAnimation', [
                      _Field('type', default='animation'),
                      'media',
                      'thumb',
                      'caption',
                      'parse_mode',
                      'width',
                      'height',
                      'duration'
                  ])

# outgoing
InputMediaAudio = _create_class('InputMediaAudio', [
                      _Field('type', default='audio'),
                      'media',
                      'thumb',
                      'caption',
                      'parse_mode',
                      'duration',
                      'performer',
                      'title'
                  ])

# outgoing
InputMediaDocument = _create_class('InputMediaDocument', [
                      _Field('type', default='document'),
                      'media',
                      'thumb',
                      'caption',
                      'parse_mode'
                  ])

# incoming
ResponseParameters = _create_class('ResponseParameters', [
                         'migrate_to_chat_id',
                         'retry_after',
                     ])



# coding=utf8

import time
import threading
import pprint
import sys
import traceback
import amanobot
import amanobot.namedtuple

"""
This script tests:
- receiving all types of messages, by asking user to produce each

Run it by:
$ python2 test.py <token> <user_id>

It will assume the bot identified by <token>, and only communicate with the user identified by <user_id>.

If you don't know your user id, run:
$ python test.py <token> 0

And send it a message anyway. It will print out your user id as an unauthorized user.
Ctrl-C to kill it, then run the proper command again.
"""

def equivalent(data, nt):
    if type(data) is dict:
        keys = data.keys()

        # number of dictionary keys == number of non-None values in namedtuple?
        if len(keys) != len([f for f in nt._fields if getattr(nt, f) is not None]):
            return False

        # map `from` to `from_`
        fields = list(map(lambda k: k+'_' if k in ['from'] else k, keys))

        return all(map(equivalent, [data[k] for k in keys], [getattr(nt, f) for f in fields]))
    elif type(data) is list:
        return all(map(equivalent, data, nt))
    else:
        return data==nt

def examine(result, type):
    try:
        print 'Examining %s ......' % type

        nt = type(**result)
        assert equivalent(result, nt), 'Not equivalent:::::::::::::::\n%s\n::::::::::::::::\n%s' % (result, nt)

        pprint.pprint(result)
        pprint.pprint(nt)
        print
    except AssertionError:
        traceback.print_exc()
        answer = raw_input('Do you want to continue? [y] ')
        if answer != 'y':
            exit(1)

expected_content_type = None
content_type_iterator = iter([
    'text', 'voice', 'sticker', 'photo', 'audio' ,'document', 'video', 'contact', 'location',
    'new_chat_member',  'new_chat_title', 'new_chat_photo',  'delete_chat_photo', 'left_chat_member'
])

def see_every_content_types(msg):
    global expected_content_type, content_type_iterator

    flavor = amanobot.flavor(msg)

    if flavor == 'chat':
        content_type, chat_type, chat_id = amanobot.glance(msg)
        from_id = msg['from']['id']

        if chat_id != USER_ID and from_id != USER_ID:
            print 'Unauthorized user:', chat_id, from_id
            return

        examine(msg, amanobot.namedtuple.Message)
        try:
            if content_type == expected_content_type:
                expected_content_type = content_type_iterator.next()
                bot.sendMessage(chat_id, 'Please give me a %s.' % expected_content_type)
            else:
                bot.sendMessage(chat_id, 'It is not a %s. Please give me a %s, please.' % (expected_content_type, expected_content_type))
        except StopIteration:
            # reply to sender because I am kicked from group already
            bot.sendMessage(from_id, 'Thank you. I am done.')

    else:
        raise amanobot.BadFlavor(msg)


TOKEN = sys.argv[1]
USER_ID = long(sys.argv[2])

bot = amanobot.Bot(TOKEN)

expected_content_type = content_type_iterator.next()
bot.sendMessage(USER_ID, 'Please give me a %s.' % expected_content_type)

bot.message_loop(see_every_content_types, run_forever=True)



import asyncio
import aiohttp
import async_timeout
import atexit
import re
import json
from .. import exception
from ..api import _methodurl, _which_pool, _fileurl, _guess_filename

_loop = asyncio.get_event_loop()

_pools = {
    'default': aiohttp.ClientSession(
                   connector=aiohttp.TCPConnector(limit=10),
                   loop=_loop)
}

_timeout = 30
_proxy = None  # (url, (username, password))


def set_proxy(url, basic_auth=None):
    global _proxy
    if not url:
        _proxy = None
    else:
        _proxy = (url, basic_auth) if basic_auth else (url,)


def _proxy_kwargs():
    if _proxy is None or len(_proxy) == 0:
        return {}
    elif len(_proxy) == 1:
        return {'proxy': _proxy[0]}
    elif len(_proxy) == 2:
        return {'proxy': _proxy[0], 'proxy_auth': aiohttp.BasicAuth(*_proxy[1])}
    else:
        raise RuntimeError("_proxy has invalid length")


async def _close_pools():
    global _pools
    for s in _pools.values():
        await s.close()

atexit.register(lambda: _loop.create_task(_close_pools()))  # have to wrap async function


def _create_onetime_pool():
    return aiohttp.ClientSession(
               connector=aiohttp.TCPConnector(limit=1, force_close=True),
               loop=_loop)


def _default_timeout(req, **user_kw):
    return _timeout


def _compose_timeout(req, **user_kw):
    token, method, params, files = req

    if method == 'getUpdates' and params and 'timeout' in params:
        # Ensure HTTP timeout is longer than getUpdates timeout
        return params['timeout'] + _default_timeout(req, **user_kw)
    elif files:
        # Disable timeout if uploading files. For some reason, the larger the file,
        # the longer it takes for the server to respond (after upload is finished).
        # It is unclear how long timeout should be.
        return None
    else:
        return _default_timeout(req, **user_kw)


def _compose_data(req, **user_kw):
    token, method, params, files = req

    data = aiohttp.FormData()

    if params:
        for key,value in params.items():
            data.add_field(key, str(value))

    if files:
        for key,f in files.items():
            if isinstance(f, tuple):
                if len(f) == 2:
                    filename, fileobj = f
                else:
                    raise ValueError('Tuple must have exactly 2 elements: filename, fileobj')
            else:
                filename, fileobj = _guess_filename(f) or key, f

            data.add_field(key, fileobj, filename=filename)

    return data


def _transform(req, **user_kw):
    timeout = _compose_timeout(req, **user_kw)

    data = _compose_data(req, **user_kw)

    url = _methodurl(req, **user_kw)

    name = _which_pool(req, **user_kw)

    if name is None:
        session = _create_onetime_pool()
        cleanup = session.close  # one-time session: remember to close
    else:
        session = _pools[name]
        cleanup = None  # reuse: do not close

    kwargs = {'data':data}
    kwargs.update(user_kw)

    return session.post, (url,), kwargs, timeout, cleanup


async def _parse(response):
    try:
        data = await response.json()
        if data is None:
            raise ValueError()
    except (ValueError, json.JSONDecodeError, aiohttp.ClientResponseError):
        text = await response.text()
        raise exception.BadHTTPResponse(response.status, text, response)

    if data['ok']:
        return data['result']
    else:
        description, error_code = data['description'], data['error_code']

        # Look for specific error ...
        for e in exception.TelegramError.__subclasses__():
            n = len(e.DESCRIPTION_PATTERNS)
            if any(map(re.search, e.DESCRIPTION_PATTERNS, n*[description], n*[re.IGNORECASE])):
                raise e(description, error_code, data)

        # ... or raise generic error
        raise exception.TelegramError(description, error_code, data)


async def request(req, **user_kw):
    fn, args, kwargs, timeout, cleanup = _transform(req, **user_kw)

    kwargs.update(_proxy_kwargs())
    try:
        if timeout is None:
            async with fn(*args, **kwargs) as r:
                return await _parse(r)
        else:
            try:
                with async_timeout.timeout(timeout):
                    async with fn(*args, **kwargs) as r:
                        return await _parse(r)

            except asyncio.TimeoutError:
                raise exception.TelegramError('Response timeout', 504, {})

    except aiohttp.ClientConnectionError:
        raise exception.TelegramError('Connection Error', 400, {})

    finally:
        if cleanup:  # e.g. closing one-time session
            if asyncio.iscoroutinefunction(cleanup):
                await cleanup()
            else:
                cleanup()


def download(req):
    session = _create_onetime_pool()

    kwargs = {}
    kwargs.update(_proxy_kwargs())

    return session, session.get(_fileurl(req), timeout=_timeout, **kwargs)
    # Caller should close session after download is complete



import sys
import time
from pprint import pprint
import amanobot
from amanobot.namedtuple import StickerSet

TOKEN = sys.argv[1]
USER_ID = int(sys.argv[2])
STICKER_SET = sys.argv[3]

bot = amanobot.Bot(TOKEN)

f = bot.uploadStickerFile(USER_ID, open('gandhi.png', 'rb'))
print('Uploaded Gandhi')

bot.addStickerToSet(USER_ID, STICKER_SET, f['file_id'], '\U0001f60a')
bot.addStickerToSet(USER_ID, STICKER_SET, open('lincoln.png', 'rb'), '\U0001f60a')
print('Added Gandhi and Lincoln to set')

s = bot.getStickerSet(STICKER_SET)
pprint(s)

ss = StickerSet(**s)

for s in ss.stickers:
    bot.deleteStickerFromSet(s.file_id)
    print('Deleted', s.file_id)
    time.sleep(3)  # throttle

s = bot.getStickerSet(STICKER_SET)
pprint(s)



import sys
import time
import amanobot
import amanobot.namedtuple
from amanobot.routing import by_content_type, make_content_type_routing_table
from amanobot.exception import NotEnoughRightsError

class AdminBot(amanobot.Bot):
    def on_chat_message(self, msg):
        content_type, chat_type, chat_id = amanobot.glance(msg)

        if 'edit_date' not in msg:
            self.sendMessage(chat_id, 'Edit the message, please.')
        else:
            self.sendMessage(chat_id, 'Add me to a group, please.')
            r = amanobot.helper.Router(by_content_type(), make_content_type_routing_table(self))
            self._router.routing_table['chat'] = r.route

    def on_new_chat_member(self, msg, new_chat_member):
        print 'New chat member:', new_chat_member
        content_type, chat_type, chat_id = amanobot.glance(msg)

        r = self.getChat(chat_id)
        print r

        r = self.getChatAdministrators(chat_id)
        print r
        print amanobot.namedtuple.ChatMemberArray(r)

        r = self.getChatMembersCount(chat_id)
        print r

        while 1:
            try:
                self.setChatTitle(chat_id, 'AdminBot Title')
                print 'Set title successfully.'
                break
            except NotEnoughRightsError:
                print 'No right to set title. Try again in 10 seconds ...'
                time.sleep(10)

        while 1:
            try:
                self.setChatPhoto(chat_id, open('gandhi.png', 'rb'))
                print 'Set photo successfully.'
                time.sleep(2)  # let tester see photo briefly
                break
            except NotEnoughRightsError:
                print 'No right to set photo. Try again in 10 seconds ...'
                time.sleep(10)

        while 1:
            try:
                self.deleteChatPhoto(chat_id)
                print 'Delete photo successfully.'
                break
            except NotEnoughRightsError:
                print 'No right to delete photo. Try again in 10 seconds ...'
                time.sleep(10)

        print 'I am done. You may remove me from the group.'


TOKEN = sys.argv[1]

bot = AdminBot(TOKEN)
bot.message_loop()
print 'Send me a text message ...'

while 1:
    time.sleep(1)



import sys
import time
from pprint import pprint
import amanobot
from amanobot.namedtuple import (
    LabeledPrice, Invoice, PreCheckoutQuery, ShippingQuery, ShippingOption,
    SuccessfulPayment)
from amanobot.loop import MessageLoop

"""
This script tests the payment process:
1. Send an invoice
2. Receive a shipping query, respond with answerShippingQuery()
3. Receive a pre-checkout query, respond with answerPreCheckoutQuery()
4. Receive a successful payment

Run it by:
$ python2 script.py <bot-token> <payment-provider-token>
"""

def on_chat_message(msg):
    content_type, chat_type, chat_id = amanobot.glance(msg)
    print content_type, chat_type, chat_id

    if content_type != 'successful_payment':
        sent = bot.sendInvoice(
                   chat_id, "Nick's Hand Cream", "Keep a man's hand like a woman's",
                   payload='a-string-identifying-related-payment-messages-tuvwxyz',
                   provider_token=PAYMENT_PROVIDER_TOKEN,
                   start_parameter='abc',
                   currency='HKD', prices=[
                       LabeledPrice(label='One Case', amount=987),
                       LabeledPrice(label='Package', amount=12)],
                   need_shipping_address=True, is_flexible=True)  # required for shipping query
        # 'Pay' button appears automatically

        pprint(sent)
        print Invoice(**sent['invoice'])

    else:
        print 'Successful payment RECEIVED!!!'
        pprint(msg)
        print SuccessfulPayment(**msg['successful_payment'])

def on_shipping_query(msg):
    query_id, from_id, invoice_payload = amanobot.glance(msg, flavor='shipping_query')

    print 'Shipping query:'
    print query_id, from_id, invoice_payload
    pprint(msg)
    print ShippingQuery(**msg)

    bot.answerShippingQuery(
        query_id, True,
        shipping_options=[
            ShippingOption(id='fedex', title='FedEx', prices=[
                LabeledPrice(label='Local', amount=345),
                LabeledPrice(label='International', amount=2345)]),
            ShippingOption(id='dhl', title='DHL', prices=[
                LabeledPrice(label='Local', amount=342),
                LabeledPrice(label='International', amount=1234)])])

def on_pre_checkout_query(msg):
    query_id, from_id, invoice_payload, currency, total_amount = amanobot.glance(msg, flavor='pre_checkout_query', long=True)

    print 'Pre-Checkout query:'
    print query_id, from_id, invoice_payload, currency, total_amount
    pprint(msg)
    print PreCheckoutQuery(**msg)

    bot.answerPreCheckoutQuery(query_id, True)

TOKEN = sys.argv[1]
PAYMENT_PROVIDER_TOKEN = sys.argv[2]

bot = amanobot.Bot(TOKEN)
MessageLoop(bot, {'chat': on_chat_message,
                  'shipping_query': on_shipping_query,
                  'pre_checkout_query': on_pre_checkout_query}).run_as_thread()

while 1:
    time.sleep(10)



"""
Like :mod:`amanobot.delegate`, this module has a bunch of seeder factories
and delegator factories.

.. autofunction:: per_chat_id
.. autofunction:: per_chat_id_in
.. autofunction:: per_chat_id_except
.. autofunction:: per_from_id
.. autofunction:: per_from_id_in
.. autofunction:: per_from_id_except
.. autofunction:: per_inline_from_id
.. autofunction:: per_inline_from_id_in
.. autofunction:: per_inline_from_id_except
.. autofunction:: per_application
.. autofunction:: per_message
.. autofunction:: per_event_source_id
.. autofunction:: per_callback_query_chat_id
.. autofunction:: per_callback_query_origin
.. autofunction:: per_invoice_payload
.. autofunction:: until
.. autofunction:: chain
.. autofunction:: pair
.. autofunction:: pave_event_space
.. autofunction:: include_callback_query_chat_id
.. autofunction:: intercept_callback_query_origin
"""

import asyncio
import traceback
from .. import exception
from . import helper

# Mirror traditional version to avoid having to import one more module
from ..delegate import (
    per_chat_id, per_chat_id_in, per_chat_id_except,
    per_from_id, per_from_id_in, per_from_id_except,
    per_inline_from_id, per_inline_from_id_in, per_inline_from_id_except,
    per_application, per_message, per_event_source_id,
    per_callback_query_chat_id, per_callback_query_origin, per_invoice_payload,
    until, chain, pair, pave_event_space,
    include_callback_query_chat_id, intercept_callback_query_origin
)

def _ensure_coroutine_function(fn):
    return fn if asyncio.iscoroutinefunction(fn) else asyncio.coroutine(fn)

def call(corofunc, *args, **kwargs):
    """
    :return:
        a delegator function that returns a coroutine object by calling
        ``corofunc(seed_tuple, *args, **kwargs)``.
    """
    corofunc = _ensure_coroutine_function(corofunc)
    def f(seed_tuple):
        return corofunc(seed_tuple, *args, **kwargs)
    return f

def create_run(cls, *args, **kwargs):
    """
    :return:
        a delegator function that calls the ``cls`` constructor whose arguments being
        a seed tuple followed by supplied ``*args`` and ``**kwargs``, then returns
        a coroutine object by calling the object's ``run`` method, which should be
        a coroutine function.
    """
    def f(seed_tuple):
        j = cls(seed_tuple, *args, **kwargs)
        return _ensure_coroutine_function(j.run)()
    return f

def create_open(cls, *args, **kwargs):
    """
    :return:
        a delegator function that calls the ``cls`` constructor whose arguments being
        a seed tuple followed by supplied ``*args`` and ``**kwargs``, then returns
        a looping coroutine object that uses the object's ``listener`` to wait for
        messages and invokes instance method ``open``, ``on_message``, and ``on_close``
        accordingly.
    """
    def f(seed_tuple):
        j = cls(seed_tuple, *args, **kwargs)

        async def wait_loop():
            bot, msg, seed = seed_tuple
            try:
                handled = await helper._invoke(j.open, msg, seed)
                if not handled:
                    await helper._invoke(j.on_message, msg)

                while 1:
                    msg = await j.listener.wait()
                    await helper._invoke(j.on_message, msg)

            # These exceptions are "normal" exits.
            except (exception.IdleTerminate, exception.StopListening) as e:
                await helper._invoke(j.on_close, e)

            # Any other exceptions are accidents. **Print it out.**
            # This is to prevent swallowing exceptions in the case that on_close()
            # gets overridden but fails to account for unexpected exceptions.
            except Exception as e:
                traceback.print_exc()
                await helper._invoke(j.on_close, e)

        return wait_loop()
    return f



import sys
import time
import amanobot
import amanobot.namedtuple
from amanobot.routing import by_content_type, make_content_type_routing_table
from amanobot.exception import NotEnoughRightsError

class AdminBot(amanobot.Bot):
    def on_chat_message(self, msg):
        content_type, chat_type, chat_id = amanobot.glance(msg)

        if 'edit_date' not in msg:
            self.sendMessage(chat_id, 'Edit the message, please.')
        else:
            self.sendMessage(chat_id, 'Add me to a group, please.')

            # Make a router to route `new_chat_member` and `left_chat_member`
            r = amanobot.helper.Router(by_content_type(), make_content_type_routing_table(self))

            # Replace current handler with that router
            self._router.routing_table['chat'] = r.route

    def on_new_chat_member(self, msg, new_chat_member):
        print('New chat member:', new_chat_member)
        content_type, chat_type, chat_id = amanobot.glance(msg)

        r = self.getChat(chat_id)
        print(r)

        r = self.getChatAdministrators(chat_id)
        print(r)
        print(amanobot.namedtuple.ChatMemberArray(r))

        r = self.getChatMembersCount(chat_id)
        print(r)

        while 1:
            try:
                self.setChatTitle(chat_id, 'AdminBot Title')
                print('Set title successfully.')
                break
            except NotEnoughRightsError:
                print('No right to set title. Try again in 10 seconds ...')
                time.sleep(10)

        while 1:
            try:
                self.setChatPhoto(chat_id, open('gandhi.png', 'rb'))
                print('Set photo successfully.')
                time.sleep(2)  # let tester see photo briefly
                break
            except NotEnoughRightsError:
                print('No right to set photo. Try again in 10 seconds ...')
                time.sleep(10)

        while 1:
            try:
                self.deleteChatPhoto(chat_id)
                print('Delete photo successfully.')
                break
            except NotEnoughRightsError:
                print('No right to delete photo. Try again in 10 seconds ...')
                time.sleep(10)

        print('I am done. Remove me from the group.')

    def on_left_chat_member(self, msg, left_chat_member):
        print('I see that I have left.')


TOKEN = sys.argv[1]

bot = AdminBot(TOKEN)
bot.message_loop()
print('Send me a text message ...')

while 1:
    time.sleep(1)



import sys
import time
import random
import amanobot.helper
from amanobot.routing import (by_content_type, make_content_type_routing_table,
                             lower_key, by_chat_command, make_routing_table,
                             by_regex)

def random_key(msg):
    return random.choice([
        0,
        (1,),
        (2, ('a1',)),
        (3, ('a1', 'a2'), {'b1': 'b'}),
        (4, (), {'kw4': 4444, 'kw5': 'xyz'}),
        ((None,), ()),
    ])

def zero(msg):
    print('Zero')

def one(msg):
    print('One')

def two(msg, a1):
    print('Two', a1)

def three(msg, a1, a2, b1):
    print('Three', a1, a2, b1)

def none_tuple(msg):
    print('None tuple')

def none_of_above(msg, *args, **kwargs):
    print('None of above', msg, args, kwargs)

top_router = amanobot.helper.Router(random_key, {0: zero,
                                                1: one,
                                                2: two,
                                                3: three,
                                                (None,): none_tuple,
                                                None: none_of_above})

for i in range(0,20):
    top_router.route({})
print()


class ContentTypeHandler(object):
    def on_text(self, msg, text):
        print('Text', msg, text)

    def on_photo(self, msg, photo):
        print('Photo', msg, photo)

def make_message_like(mm):
    for d in mm:
        d.update({'chat': {'type': 'private', 'id': 1000}})

top_router.key_function = by_content_type()
top_router.routing_table = make_content_type_routing_table(ContentTypeHandler())
del top_router.routing_table['video']  # let video fall to default handler
top_router.routing_table[None] = none_of_above

messages = [{'text': 'abc'},
            {'photo': 'some photo'},
            {'video': 'some video'},]
make_message_like(messages)

for i in range(0,10):
    top_router.route(random.choice(messages))
print()


class CommandHandler(object):
    def on_start(self, msg):
        print('Command: start', msg)

    def on_settings(self, msg):
        print('Command: settings', msg)

    def on_invalid_text(self, msg):
        print('Invalid text', msg)

    def on_invalid_command(self, msg):
        print('Invalid command', msg)

command_handler = CommandHandler()
command_router = amanobot.helper.Router(lower_key(by_chat_command()),
                                       make_routing_table(command_handler, [
                                           'start',
                                           'settings',
                                           ((None,), command_handler.on_invalid_text),
                                           (None, command_handler.on_invalid_command),
                                       ]))

top_router.routing_table['text'] = command_router.route

messages = [{'text': '/start'},
            {'text': '/SETTINGS'},
            {'text': '/bad'},
            {'text': 'plain text'},
            {'photo': 'some photo'},
            {'video': 'some video'},]
make_message_like(messages)

for i in range(0,20):
    top_router.route(random.choice(messages))
print()


class RegexHandler(object):
    def on_CS101(self, msg, match):
        print('Someone mentioned CS101 !!!', msg, match.groups())

    def on_CS202(self, msg, match):
        print('Someone mentioned CS202 !!!', msg, match.groups())

    def no_cs_courses_mentioned(self, msg):
        print('No CS courses mentioned ...', msg)

    def course_not_exist(self, msg, match):
        print('%s does not exist' % match.group(1), msg)

regex_handler = RegexHandler()
regex_router = amanobot.helper.Router(by_regex(lambda msg: msg['text'], '(CS[0-9]{3})'),
                                     make_routing_table(regex_handler, [
                                         'CS101',
                                         'CS202',
                                         ((None,), regex_handler.no_cs_courses_mentioned),
                                         (None, regex_handler.course_not_exist),
                                     ]))

command_router.routing_table[(None,)] = regex_router.route

messages = [{'text': '/start'},
            {'text': '/SETTINGS'},
            {'text': '/bad'},
            {'text': 'plain text'},
            {'text': 'I want to take CS101.'},
            {'text': 'I\'d rather take CS202.'},
            {'text': 'Why don\'t you take CS303?'},
            {'text': 'I hate computer science!'},
            {'photo': 'some photo'},
            {'video': 'some video'},]
make_message_like(messages)

for i in range(0,30):
    top_router.route(random.choice(messages))
print()


TOKEN = sys.argv[1]

bot = amanobot.Bot(TOKEN)
bot._router.routing_table['chat'] = top_router.route

bot.message_loop()
print('Send me some messages ...')

while 1:
    time.sleep(10)



import urllib3
import logging
import sys
import json
import re
import os

from . import exception, _isstring

# Suppress InsecurePlatformWarning
urllib3.disable_warnings()


PY_3 = sys.version_info.major >= 3

_default_pool_params = dict(num_pools=3, maxsize=10, retries=3, timeout=30)
_onetime_pool_params = dict(num_pools=1, maxsize=1, retries=3, timeout=30)

_pools = {
    'default': urllib3.PoolManager(**_default_pool_params),
}

_onetime_pool_spec = (urllib3.PoolManager, _onetime_pool_params)


def set_proxy(url, basic_auth=None):
    """
    Access Bot API through a proxy.

    :param url: proxy URL
    :param basic_auth: 2-tuple ``('username', 'password')``
    """
    global _pools, _onetime_pool_spec
    if not url:
        _pools['default'] = urllib3.PoolManager(**_default_pool_params)
        _onetime_pool_spec = (urllib3.PoolManager, _onetime_pool_params)
    elif basic_auth:
        h = urllib3.make_headers(proxy_basic_auth=':'.join(basic_auth))
        _pools['default'] = urllib3.ProxyManager(url, proxy_headers=h, **_default_pool_params)
        _onetime_pool_spec = (urllib3.ProxyManager, dict(proxy_url=url, proxy_headers=h, **_onetime_pool_params))
    else:
        _pools['default'] = urllib3.ProxyManager(url, **_default_pool_params)
        _onetime_pool_spec = (urllib3.ProxyManager, dict(proxy_url=url, **_onetime_pool_params))


def _create_onetime_pool():
    cls, kw = _onetime_pool_spec
    return cls(**kw)


def _methodurl(req, **user_kw):
    token, method, params, files = req
    return 'https://api.telegram.org/bot%s/%s' % (token, method)


def _which_pool(req, **user_kw):
    token, method, params, files = req
    return None if files else 'default'


def _guess_filename(obj):
    name = getattr(obj, 'name', None)
    if name and _isstring(name) and name[0] != '<' and name[-1] != '>':
        return os.path.basename(name)


def _filetuple(key, f):
    if not isinstance(f, tuple):
        return (_guess_filename(f) or key, f.read())
    elif len(f) == 1:
        return (_guess_filename(f[0]) or key, f[0].read())
    elif len(f) == 2:
        return (f[0], f[1].read())
    elif len(f) == 3:
        return (f[0], f[1].read(), f[2])
    else:
        raise ValueError()


def _fix_type(v):
    if isinstance(v, float if PY_3 else (long, float)):
        return str(v)
    else:
        return v


def _compose_fields(req, **user_kw):
    token, method, params, files = req

    fields = {k:_fix_type(v) for k,v in params.items()} if params is not None else {}
    if files:
        fields.update({k:_filetuple(k,v) for k,v in files.items()})

    return fields


def _default_timeout(req, **user_kw):
    name = _which_pool(req, **user_kw)
    if name is None:
        return _onetime_pool_spec[1]['timeout']
    else:
        return _pools[name].connection_pool_kw['timeout']


def _compose_kwargs(req, **user_kw):
    token, method, params, files = req
    kw = {}

    if not params and not files:
        kw['encode_multipart'] = False

    if method == 'getUpdates' and params and 'timeout' in params:
        # Ensure HTTP timeout is longer than getUpdates timeout
        kw['timeout'] = params['timeout'] + _default_timeout(req, **user_kw)
    elif files:
        # Disable timeout if uploading files. For some reason, the larger the file,
        # the longer it takes for the server to respond (after upload is finished).
        # It is unclear how long timeout should be.
        kw['timeout'] = None

    # Let user-supplied arguments override
    kw.update(user_kw)
    return kw


def _transform(req, **user_kw):
    kwargs = _compose_kwargs(req, **user_kw)

    fields = _compose_fields(req, **user_kw)

    url = _methodurl(req, **user_kw)

    name = _which_pool(req, **user_kw)

    if name is None:
        pool = _create_onetime_pool()
    else:
        pool = _pools[name]

    return pool.request_encode_body, ('POST', url, fields), kwargs


def _parse(response):
    try:
        text = response.data.decode('utf-8')
        data = json.loads(text)
    except ValueError:  # No JSON object could be decoded
        raise exception.BadHTTPResponse(response.status, text, response)

    if data['ok']:
        return data['result']
    else:
        description, error_code = data['description'], data['error_code']

        # Look for specific error ...
        for e in exception.TelegramError.__subclasses__():
            n = len(e.DESCRIPTION_PATTERNS)
            if any(map(re.search, e.DESCRIPTION_PATTERNS, n*[description], n*[re.IGNORECASE])):
                raise e(description, error_code, data)

        # ... or raise generic error
        raise exception.TelegramError(description, error_code, data)


def request(req, **user_kw):
    fn, args, kwargs = _transform(req, **user_kw)
    r = fn(*args, **kwargs)  # `fn` must be thread-safe
    return _parse(r)


def _fileurl(req):
    token, path = req
    return 'https://api.telegram.org/file/bot%s/%s' % (token, path)


def download(req, **user_kw):
    pool = _create_onetime_pool()
    r = pool.request('GET', _fileurl(req), **user_kw)
    return r



import asyncio
import traceback
from .. import filtering, helper, exception
from .. import (
    flavor, chat_flavors, inline_flavors, is_event,
    message_identifier, origin_identifier)

# Mirror traditional version
from ..helper import (
    Sender, Administrator, Editor, openable,
    StandardEventScheduler, StandardEventMixin)


async def _invoke(fn, *args, **kwargs):
    if asyncio.iscoroutinefunction(fn):
        return await fn(*args, **kwargs)
    else:
        return fn(*args, **kwargs)


def _create_invoker(obj, method_name):
    async def d(*a, **kw):
        method = getattr(obj, method_name)
        return await _invoke(method, *a, **kw)
    return d


class Microphone(object):
    def __init__(self):
        self._queues = set()

    def add(self, q):
        self._queues.add(q)

    def remove(self, q):
        self._queues.remove(q)

    def send(self, msg):
        for q in self._queues:
            try:
                q.put_nowait(msg)
            except asyncio.QueueFull:
                traceback.print_exc()
                pass


class Listener(helper.Listener):
    async def wait(self):
        """
        Block until a matched message appears.
        """
        if not self._patterns:
            raise RuntimeError('Listener has nothing to capture')

        while 1:
            msg = await self._queue.get()

            if any(map(lambda p: filtering.match_all(msg, p), self._patterns)):
                return msg


from concurrent.futures._base import CancelledError

class Answerer(object):
    """
    When processing inline queries, ensures **at most one active task** per user id.
    """

    def __init__(self, bot, loop=None):
        self._bot = bot
        self._loop = loop if loop is not None else asyncio.get_event_loop()
        self._working_tasks = {}

    def answer(self, inline_query, compute_fn, *compute_args, **compute_kwargs):
        """
        Create a task that calls ``compute fn`` (along with additional arguments
        ``*compute_args`` and ``**compute_kwargs``), then applies the returned value to
        :meth:`.Bot.answerInlineQuery` to answer the inline query.
        If a preceding task is already working for a user, that task is cancelled,
        thus ensuring at most one active task per user id.

        :param inline_query:
            The inline query to be processed. The originating user is inferred from ``msg['from']['id']``.

        :param compute_fn:
            A function whose returned value is given to :meth:`.Bot.answerInlineQuery` to send.
            May return:

            - a *list* of `InlineQueryResult <https://core.telegram.org/bots/api#inlinequeryresult>`_
            - a *tuple* whose first element is a list of `InlineQueryResult <https://core.telegram.org/bots/api#inlinequeryresult>`_,
              followed by positional arguments to be supplied to :meth:`.Bot.answerInlineQuery`
            - a *dictionary* representing keyword arguments to be supplied to :meth:`.Bot.answerInlineQuery`

        :param \*compute_args: positional arguments to ``compute_fn``
        :param \*\*compute_kwargs: keyword arguments to ``compute_fn``
        """

        from_id = inline_query['from']['id']

        async def compute_and_answer():
            try:
                query_id = inline_query['id']

                ans = await _invoke(compute_fn, *compute_args, **compute_kwargs)

                if isinstance(ans, list):
                    await self._bot.answerInlineQuery(query_id, ans)
                elif isinstance(ans, tuple):
                    await self._bot.answerInlineQuery(query_id, *ans)
                elif isinstance(ans, dict):
                    await self._bot.answerInlineQuery(query_id, **ans)
                else:
                    raise ValueError('Invalid answer format')
            except CancelledError:
                # Cancelled. Record has been occupied by new task. Don't touch.
                raise
            except:
                # Die accidentally. Remove myself from record.
                del self._working_tasks[from_id]
                raise
            else:
                # Die naturally. Remove myself from record.
                del self._working_tasks[from_id]

        if from_id in self._working_tasks:
            self._working_tasks[from_id].cancel()

        t = self._loop.create_task(compute_and_answer())
        self._working_tasks[from_id] = t


class AnswererMixin(helper.AnswererMixin):
    Answerer = Answerer  # use async Answerer class


class CallbackQueryCoordinator(helper.CallbackQueryCoordinator):
    def augment_send(self, send_func):
        async def augmented(*aa, **kw):
            sent = await send_func(*aa, **kw)

            if self._enable_chat and self._contains_callback_data(kw):
                self.capture_origin(message_identifier(sent))

            return sent
        return augmented

    def augment_edit(self, edit_func):
        async def augmented(msg_identifier, *aa, **kw):
            edited = await edit_func(msg_identifier, *aa, **kw)

            if (edited is True and self._enable_inline) or (isinstance(edited, dict) and self._enable_chat):
                if self._contains_callback_data(kw):
                    self.capture_origin(msg_identifier)
                else:
                    self.uncapture_origin(msg_identifier)

            return edited
        return augmented

    def augment_delete(self, delete_func):
        async def augmented(msg_identifier, *aa, **kw):
            deleted = await delete_func(msg_identifier, *aa, **kw)

            if deleted is True:
                self.uncapture_origin(msg_identifier)

            return deleted
        return augmented

    def augment_on_message(self, handler):
        async def augmented(msg):
            if (self._enable_inline
                    and flavor(msg) == 'chosen_inline_result'
                    and 'inline_message_id' in msg):
                inline_message_id = msg['inline_message_id']
                self.capture_origin(inline_message_id)

            return await _invoke(handler, msg)
        return augmented


class InterceptCallbackQueryMixin(helper.InterceptCallbackQueryMixin):
    CallbackQueryCoordinator = CallbackQueryCoordinator


class IdleEventCoordinator(helper.IdleEventCoordinator):
    def augment_on_message(self, handler):
        async def augmented(msg):
            # Reset timer if this is an external message
            is_event(msg) or self.refresh()
            return await _invoke(handler, msg)
        return augmented

    def augment_on_close(self, handler):
        async def augmented(ex):
            try:
                if self._timeout_event:
                    self._scheduler.cancel(self._timeout_event)
                    self._timeout_event = None
            # This closing may have been caused by my own timeout, in which case
            # the timeout event can no longer be found in the scheduler.
            except exception.EventNotFound:
                self._timeout_event = None
            return await _invoke(handler, ex)
        return augmented


class IdleTerminateMixin(helper.IdleTerminateMixin):
    IdleEventCoordinator = IdleEventCoordinator


class Router(helper.Router):
    async def route(self, msg, *aa, **kw):
        """
        Apply key function to ``msg`` to obtain a key, look up routing table
        to obtain a handler function, then call the handler function with
        positional and keyword arguments, if any is returned by the key function.

        ``*aa`` and ``**kw`` are dummy placeholders for easy nesting.
        Regardless of any number of arguments returned by the key function,
        multi-level routing may be achieved like this::

            top_router.routing_table['key1'] = sub_router1.route
            top_router.routing_table['key2'] = sub_router2.route
        """
        k = self.key_function(msg)

        if isinstance(k, (tuple, list)):
            key, args, kwargs = {1: tuple(k) + ((),{}),
                                 2: tuple(k) + ({},),
                                 3: tuple(k),}[len(k)]
        else:
            key, args, kwargs = k, (), {}

        try:
            fn = self.routing_table[key]
        except KeyError as e:
            # Check for default handler, key=None
            if None in self.routing_table:
                fn = self.routing_table[None]
            else:
                raise RuntimeError('No handler for key: %s, and default handler not defined' % str(e.args))

        return await _invoke(fn, msg, *args, **kwargs)


class DefaultRouterMixin(object):
    def __init__(self, *args, **kwargs):
        self._router = Router(flavor, {'chat': _create_invoker(self, 'on_chat_message'),
                                       'callback_query': _create_invoker(self, 'on_callback_query'),
                                       'inline_query': _create_invoker(self, 'on_inline_query'),
                                       'chosen_inline_result': _create_invoker(self, 'on_chosen_inline_result'),
                                       'shipping_query': _create_invoker(self, 'on_shipping_query'),
                                       'pre_checkout_query': _create_invoker(self, 'on_pre_checkout_query'),
                                       '_idle': _create_invoker(self, 'on__idle')})

        super(DefaultRouterMixin, self).__init__(*args, **kwargs)

    @property
    def router(self):
        """ See :class:`.helper.Router` """
        return self._router

    async def on_message(self, msg):
        """
        Called when a message is received.
        By default, call :meth:`Router.route` to handle the message.
        """
        await self._router.route(msg)


@openable
class Monitor(helper.ListenerContext, DefaultRouterMixin):
    def __init__(self, seed_tuple, capture, **kwargs):
        """
        A delegate that never times-out, probably doing some kind of background monitoring
        in the application. Most naturally paired with :func:`amanobot.aio.delegate.per_application`.

        :param capture: a list of patterns for ``listener`` to capture
        """
        bot, initial_msg, seed = seed_tuple
        super(Monitor, self).__init__(bot, seed, **kwargs)

        for pattern in capture:
            self.listener.capture(pattern)


@openable
class ChatHandler(helper.ChatContext,
                  DefaultRouterMixin,
                  StandardEventMixin,
                  IdleTerminateMixin):
    def __init__(self, seed_tuple,
                 include_callback_query=False, **kwargs):
        """
        A delegate to handle a chat.
        """
        bot, initial_msg, seed = seed_tuple
        super(ChatHandler, self).__init__(bot, seed, **kwargs)

        self.listener.capture([{'chat': {'id': self.chat_id}}])

        if include_callback_query:
            self.listener.capture([{'message': {'chat': {'id': self.chat_id}}}])


@openable
class UserHandler(helper.UserContext,
                  DefaultRouterMixin,
                  StandardEventMixin,
                  IdleTerminateMixin):
    def __init__(self, seed_tuple,
                 include_callback_query=False,
                 flavors=chat_flavors+inline_flavors, **kwargs):
        """
        A delegate to handle a user's actions.

        :param flavors:
            A list of flavors to capture. ``all`` covers all flavors.
        """
        bot, initial_msg, seed = seed_tuple
        super(UserHandler, self).__init__(bot, seed, **kwargs)

        if flavors == 'all':
            self.listener.capture([{'from': {'id': self.user_id}}])
        else:
            self.listener.capture([lambda msg: flavor(msg) in flavors, {'from': {'id': self.user_id}}])

        if include_callback_query:
            self.listener.capture([{'message': {'chat': {'id': self.user_id}}}])


class InlineUserHandler(UserHandler):
    def __init__(self, seed_tuple, **kwargs):
        """
        A delegate to handle a user's inline-related actions.
        """
        super(InlineUserHandler, self).__init__(seed_tuple, flavors=inline_flavors, **kwargs)


@openable
class CallbackQueryOriginHandler(helper.CallbackQueryOriginContext,
                                 DefaultRouterMixin,
                                 StandardEventMixin,
                                 IdleTerminateMixin):
    def __init__(self, seed_tuple, **kwargs):
        """
        A delegate to handle callback query from one origin.
        """
        bot, initial_msg, seed = seed_tuple
        super(CallbackQueryOriginHandler, self).__init__(bot, seed, **kwargs)

        self.listener.capture([
            lambda msg:
                flavor(msg) == 'callback_query' and origin_identifier(msg) == self.origin
        ])


@openable
class InvoiceHandler(helper.InvoiceContext,
                     DefaultRouterMixin,
                     StandardEventMixin,
                     IdleTerminateMixin):
    def __init__(self, seed_tuple, **kwargs):
        """
        A delegate to handle messages related to an invoice.
        """
        bot, initial_msg, seed = seed_tuple
        super(InvoiceHandler, self).__init__(bot, seed, **kwargs)

        self.listener.capture([{'invoice_payload': self.payload}])
        self.listener.capture([{'successful_payment': {'invoice_payload': self.payload}}])



def pick(obj, keys):
    def pick1(k):
        if type(obj) is dict:
            return obj[k]
        else:
            return getattr(obj, k)

    if isinstance(keys, list):
        return [pick1(k) for k in keys]
    else:
        return pick1(keys)

def match(data, template):
    if isinstance(template, dict) and isinstance(data, dict):
        def pick_and_match(kv):
            template_key, template_value = kv
            if hasattr(template_key, 'search'):  # regex
                data_keys = list(filter(template_key.search, data.keys()))
                if not data_keys:
                    return False
            elif template_key in data:
                data_keys = [template_key]
            else:
                return False
            return any(map(lambda data_value: match(data_value, template_value), pick(data, data_keys)))

        return all(map(pick_and_match, template.items()))
    elif callable(template):
        return template(data)
    else:
        return data == template

def match_all(msg, templates):
    return all(map(lambda t: match(msg, t), templates))
